import keras_tuner as kt
from model_creator.auto_encoder import Autoencoder
from model_creator.auto_encoder_tuning import Autoencoder_Tuning
from model_creator.tuner import MyTuner
import model_creator.config_default as conf
import tensorflow as tf
import numpy as np
import os

class CreateData:
    def __init__(self, species_name : list) -> None:
        self.species_name = species_name

    def load_music(self):
        """
        First load the spectrograms construct the training song set        
        """

        x_train = []
        for specie in self.species_name:
            spectrograms_path = "./preprocessed_data/" + specie + "/spectrograms"
            
            for root, _, file_names in os.walk(spectrograms_path):
                for file_name in file_names:
                    file_path = os.path.join(root, file_name)
                    spectrogram = np.load(file_path) # (n_bins, n_frames)
                    x_train.append(spectrogram)

        x_train = np.array(x_train)
        x_train = x_train[..., np.newaxis]
        np.random.shuffle(x_train)
        x_train = x_train
        
        return x_train


class ClassiqueTrain:
    """
    Build the model with default parameters adapted to birds 
    """

    def __init__(self, taille_input : tuple) -> None:
        self.taille_input = taille_input

        self.autoencoder = Autoencoder(
            input_shape=(self.taille_input[0], self.taille_input[1], 1),
            conv_filters=(1024,512, 256, 128, 64),
            conv_kernels=(3,3,3,3,2),
            conv_strides=(2,2,2,2, (2,1)),
            latent_space_dim=256,
            save_path=conf.MODEL_PATH_CLASSIQUE
        )

    def fit_classique(self, x_train : np.array, batch_size = conf.DEFAULT_BATCH_SIZE, epochs = conf.DEFAULT_EPOCHS):

        self.autoencoder.summary()
        self.autoencoder.compile(conf.DEFAULT_LEARNING_RATE)
        history = self.autoencoder.train(x_train, batch_size, epochs)
        
        return self.autoencoder, history

class ParameterTuning:
    """
    Allow model hyper-tuning
    """
    def __init__(self, dico_archi : dict, taille_input : tuple, nb_trial=conf.DEFAULT_MODEL_TUNING_TRIAL) -> None:
        self.taille_input = taille_input

        auto_tuner = Autoencoder_Tuning(input_shape=(self.taille_input[0], self.taille_input[1], 1),
            dico_param=dico_archi,
            save_path=conf.MODEL_PATH_CLASSIQUE)
        
        self.tuner = MyTuner(
            oracle = kt.oracles.RandomSearch(
                objective="loss",
                max_trials=nb_trial,
            ),
            distribution_strategy=tf.distribute.MirroredStrategy(),
            hypermodel=auto_tuner,
            overwrite=True,
            directory="my_dir/",
            project_name="hyper_tuning"
        )

    def logwandb(self, file_path:str) -> None:
        with open(file_path, 'r') as f:
            key = f.readline()
        os.environ["WANDB_API_KEY"] = key

    def tune(self, x_train, batch_size : int, epochs : int) -> None:
        self.tuner.search(x_train,batch_size = batch_size, epochs=epochs, objective = 'loss')


import numpy as np
import os
import tensorflow as tf
import keras_tuner as kt
from tensorflow.keras import Model
from tensorflow.keras.layers import Input, Conv2D, BatchNormalization, Flatten, Dense, Reshape, Conv2DTranspose, Activation, LeakyReLU, Dropout
from tensorflow.keras import backend as K
from tensorflow.keras.optimizers import Adam

tf.compat.v1.disable_eager_execution()

class Autoencoder_Tuning(kt.HyperModel):

    """
    Deep Convolutionnal autoencoder tuning with mirrored encoder and decoder components
    """

    def __init__(self, input_shape, dico_param, save_path):
        """
        Initialisation of the class with the different parameters
        """
        #Shape of the input spectrogram
        self.input_shape = input_shape # [1024,256,1]

        #Dictionnaire
        self.dico_param = dico_param

        #The save path
        self.save_path = save_path

        #Encoder model part
        self.encoder = None
        #Decoder model part
        self.decoder = None
        #The entire model encoder + decoder
        self.model = None

        self._shape_before_bottleneck = None
        self._model_input = None


    def summary(self):
        '''
        Describe the network with the summary
        '''
        self.encoder.summary()
        self.decoder.summary()
        self.model.summary()

    def compile(self, learning_rate, hp):
        """Compile the model"""

        optimizer = Adam(learning_rate  = learning_rate)
        #Weight given to the reconstruction loss
        self.reconstruction_loss_weight = 1000000

        self.model.compile(optimizer = optimizer,
                        loss = self._calculate_combined_loss, 
                        metrics = [self._calculate_reconstruction_loss, self._calculate_kl_loss])

    def _calculate_combined_loss(self, y_target, y_predicted):
        reconstruction_loss = self._calculate_reconstruction_loss(y_target, y_predicted)
        kl_loss = self._calculate_kl_loss(y_target, y_predicted)
        combined_loss = self.reconstruction_loss_weight * reconstruction_loss + kl_loss

        return combined_loss

    def _calculate_reconstruction_loss(self, y_target, y_predicted):
        error = y_target - y_predicted
        reconstruction_loss = K.mean(K.square(error), axis = [1,2,3])
        return reconstruction_loss

    
    def _calculate_kl_loss(self, y_target, y_predicted):
        kl_loss = - 0.5 * K.sum(1 + self.log_variance - K.square(self.mu) - K.exp(self.log_variance), axis=1)
        return kl_loss

    def _create_folder_if_it_doesnt_exist(self, folder):
        if not os.path.exists(folder):
            os.mkdir(folder)

    def build(self, hp):

        self.__archi_choice(hp)
        #The size of the latent space
        self.latent_space_dim = hp.Int("Size_latent _space", min_value=128, max_value=1024, step=128)
        self._build_encoder(hp)
        self._build_decoder(hp)
        self._build_autoencoder()
        self.compile(hp.Float("learning_rate_tuning", min_value=0.00005, max_value=0.0002, step=0.00005),hp)

        return self.model

    def __archi_choice(self, hp):
        """
        Permet de choisir entre differents architectures de modèle de convolution
        """
        archi_name = hp.Choice("Choix_architecture", list(self.dico_param.keys()))

        #Shape of the differents convolutials layers
        self.conv_filters = self.dico_param[archi_name]["conv_filters"]
        #Shape of the kernels
        self.conv_kernels = self.dico_param[archi_name]["conv_kernels"]
        #The different strides
        self.conv_strides = self.dico_param[archi_name]["conv_strides"]
        #Know the 
        self._num_conv_layers = len(self.conv_filters)


    def _build_autoencoder(self):
        #Input of the model
        model_input = self._model_input
        #Encoder part
        model_encoder = self.encoder(model_input)
        #Decoder part
        model_output = self.decoder(model_encoder)
        #Entire model
        self.model = Model(model_input, model_output, name="autoencoder")

    
    def _build_decoder(self, hp):
        """
        Build the different layers for the decoder
        """

        #Decoder input
        decoder_input = self._add_decoder_input()
        #Dense layer before reshaping
        dropout = hp.Boolean("Decoder_dropout")
        dense_layer = self._add_dense_layer(decoder_input, dropout=dropout)
        #Reshape layer to pass from 1D to 2D array
        reshape_layer = self._add_reshape_layer(dense_layer)
        #Add convolutionnal transpose blocks
        conv_transpose_layer = self._add_conv_transpose_layers(reshape_layer)
        #Add decoder output
        decoder_output = self._add_decoder_output(conv_transpose_layer)
        #Model final composition
        self.decoder = Model(decoder_input, decoder_output, name = 'decoder')
    
    def _add_decoder_input(self):
        return Input(shape=(self.latent_space_dim,), name="decoder_input")

    def _add_dense_layer(self, decoder_input, dropout = True):
        """
        The different blocs to decompress the botleneck
        """

        num_neurons = np.prod(self._shape_before_bottleneck)
        dense_layer = Dense(self.latent_space_dim * 2, name="decoder_dense_1", activation= 'relu')(decoder_input)

        #Permet de tester avec ou sans dropout
        if dropout:
            dense_layer = Dropout(0.3)(dense_layer)
        dense_layer = Dense(self.latent_space_dim * 4, name="decoder_dense_2", activation= 'relu')(dense_layer)

        #Permet de tester avec ou sans dropout
        if dropout:
            dense_layer = Dropout(0.3)(dense_layer)
        dense_layer = Dense(num_neurons, name="decoder_dense_3", activation = 'relu')(dense_layer)

        return dense_layer
    
    def _add_reshape_layer(self, dense_layer):
        return Reshape(self._shape_before_bottleneck)(dense_layer)

    def _add_conv_transpose_layers(self, x):
        """Add conv transpose blocks."""
        for layer_index in reversed(range(1, self._num_conv_layers)):
            x = self._add_conv_transpose_layer(layer_index, x)
        return x

    
    def _add_conv_transpose_layer(self, layer_index, x):
        layer_num = self._num_conv_layers - layer_index
        conv_transpose_layer = Conv2DTranspose(
            filters = self.conv_filters[layer_index], # [24, 24, 1]
            kernel_size = self.conv_kernels[layer_index],
            strides = self.conv_strides[layer_index],
            padding = "same",
            name=f"decoder_conv_transpose_layer_{layer_num}"
        )
        x = conv_transpose_layer(x)
        x = LeakyReLU(name=f"decoder_leaky_relu_{layer_num}")(x)
        x = BatchNormalization(name =f"decoder_bn_{layer_num}")(x)
        return x

    def _add_decoder_output(self, x):
        conv_transpose_layer = Conv2DTranspose(
            filters = 1, # [24, 24, 1]
            kernel_size = self.conv_kernels[0],
            strides = self.conv_strides[0],
            padding = "same",
            name=f"decoder_conv_transpose_layer_output{self._num_conv_layers}"
        )

        x = conv_transpose_layer(x)
        output_layer = Activation("sigmoid", name="sigmoid_layer")(x)
        return output_layer

    def _build_encoder(self, hp):
        """
        Build the different layers of the encoder
        """

        #Input
        encoder_input = self._add_encoder_input()
        #Convolution layer
        conv_layer = self._add_conv_layers(encoder_input)
        #Bottleneck with or without Dropout
        dropout = hp.Boolean("Encoder_dropout")
        bottleneck = self._add_bottleneck(conv_layer, dropout=dropout)
        self._model_input = encoder_input
        self.encoder = Model(encoder_input, bottleneck, name = "encoder")

    def _add_encoder_input(self):
        return Input(shape=self.input_shape, name = "encoder_input")
    
    def _add_conv_layers(self, encoder_input):
        x = encoder_input
        for layer_index in range(self._num_conv_layers):
            x = self._add_conv_layer(layer_index, x)
        return x
    
    def _add_conv_layer(self, layer_index, x):
        """
        Define the convolutionnal block
        """
        layer_number = layer_index + 1
        conv_layer = Conv2D(
            filters = self.conv_filters[layer_index],
            kernel_size = self.conv_kernels[layer_index],
            strides = self.conv_strides[layer_index],
            padding = "same",
            name = f"encoder_conv_layer_{layer_index + 1}"
        )

        x = conv_layer(x)
        x = LeakyReLU(name=f"encoder_leakyrelu_{layer_number}")(x)
        x = BatchNormalization(name=f"encoder_bn_{layer_number}")(x)
        return x
    
    def _add_bottleneck(self, x, dropout = True):
        """Flatten data and add bottleneck with Gaussian sampling (Dense layer)."""
        self._shape_before_bottleneck = K.int_shape(x)[1:] # [7, 7, 32]
        x = Flatten()(x)
        x = Dense(self.latent_space_dim * 4, name = "first_reduction", activation="relu")(x)
        if dropout:
            x = Dropout(0.3)(x)
        x = Dense(self.latent_space_dim * 2, name = "second_reduction", activation="relu")(x)
        if dropout:
            x = Dropout(0.3)(x)

        self.mu = Dense(self.latent_space_dim, name="mu")(x)
        self.log_variance = Dense(self.latent_space_dim, name="log_variance")(x)

        def sample_point_from_normal_distribution(args):
            mu, log_variance = args
            epsilon = K.random_normal(shape=K.shape(self.mu), mean=0., stddev=1.)
            sampled_point = mu + K.exp(log_variance / 2) * epsilon

            return sampled_point

        x = tf.keras.layers.Lambda(sample_point_from_normal_distribution, name="encoder_output")([self.mu, self.log_variance])
        return x


from model_creator.auto_encoder import Autoencoder
import numpy as np
import os

class Use_Case_Model:
    """
    Permet la construction des espaces latents moyens pour la génération de sons
    """
    
    def __init__(self, path_saved = './model'):
        self.path_saved = path_saved

    def recup_model(self):
        """
        Allow to load the model
        """

        trained_auto_encoder = Autoencoder.load(self.path_saved)
        trained_encoder = trained_auto_encoder.encoder
        trained_decoder = trained_auto_encoder.decoder

        return trained_auto_encoder, trained_encoder, trained_decoder

    def make_prediction_encoder(self, path_spectrogram, encoder) -> np.array:
        """
        Allow user to make prediction
        """

        #Load data
        example = np.load(path_spectrogram)
        #Change the dimentions
        example = np.expand_dims(example, 2)
        example = example[None,:,:,:]

        return encoder.predict(example)

    def representation_oiseaux(self, espece : str, encoder) -> np.array:
        """
        Calcule la moyenne et la variance des différents sons d'oiseau
        """

        pred = []
        for tab_spec in os.listdir("preprocessed_data/" + espece + "/spectrograms"):
            pred.append(self.make_prediction_encoder("preprocessed_data/" + espece + "/spectrograms/" + tab_spec, encoder).reshape(256))
        pred = np.array(pred)

        return np.array([np.mean(pred, axis=0), np.var(pred, axis=0)])


    def construction_utils(self, espece : list) -> None:
        """
        Permet de sauvegarder les moyennes des vecteurs
        """
        _, encoder,_ = self.recup_model()

        if not os.path.exists("model_result/"):
            os.makedirs("model_result")

        for esp in espece:
            tab = self.representation_oiseaux(esp, encoder)
            np.save(arr=tab, file="model_result/" + esp.replace(' ','_'))


import kerastuner as kt
import tensorflow as tf
import wandb
from wandb.keras import WandbCallback

class MyTuner(kt.Tuner):
    """
    Allow you to tune hyperparameters
    """

    def run_trial(self, trial, trainX, batch_size, epochs, objective):

        hp = trial.hyperparameters
        objective_name_str = objective

        ## create the model with the current trial hyperparameters
        model = self.hypermodel.build(hp)
        
        ## Initiates new run for each trial on the dashboard of Weights & Biases
        run = wandb.init(project="Birds Model Tune", config=hp.values)

        ## WandbCallback() logs all the metric data such as
        ## loss, accuracy and etc on dashboard for visualization
        history = model.fit(trainX,
                  trainX,
                  batch_size=batch_size,
                  epochs=epochs,
                  callbacks=[WandbCallback(), tf.keras.callbacks.EarlyStopping('loss', patience=5)])  

        ## if val_accurcy used, use the val_accuracy of last epoch model which is fully trained
        loss = history.history['loss'][-1]  ## [-1] will give the last value in the list

        ## Send the objective data to the oracle for comparison of hyperparameters
        self.oracle.update_trial(trial.trial_id, {objective_name_str:loss})
        
        ## ends the run on the Weights & Biases dashboard
        run.finish()


# Import the AudioSegment class for processing audio and the 
# split_on_silence function for separating out silent chunks.
import os
from pydub import AudioSegment
from pydub.silence import split_on_silence


def find_chunks(file, minimum_silence_gap:int, silence_bar:int) -> list:
    """
    Permet de trouver les différents chunks où les oiseaux chantes
    """

    # Load your audio.
    song = AudioSegment.from_mp3(file)

    # Split track where the silence is 2 seconds or more and get chunks using 
    # the imported function.
    chunks = split_on_silence (
        # Use the loaded audio.
        song, 
        # Specify that a silent chunk must be at least 2 seconds or 2000 ms long.
        min_silence_len = minimum_silence_gap,
        # Consider a chunk silent if it's quieter than -16 dBFS.
        # (You may want to adjust this parameter.)
        silence_thresh = silence_bar
    )
    return chunks

def reconstruct_chunks(chunks,exporting_dir,name,padding = 300):
    """
    Permet de reconstruire les différents chunks 
    """
    
    # Define a function to normalize a chunk to a target amplitude.
    def match_target_amplitude(aChunk, target_dBFS):
        ''' Normalize given audio chunk '''
        change_in_dBFS = target_dBFS - aChunk.dBFS
        return aChunk.apply_gain(change_in_dBFS)

    if not os.path.exists(exporting_dir):
        os.makedirs(exporting_dir)
        
    # Process each chunk with your parameters
    for i, chunk in enumerate(chunks):
        # Create a silence chunk that's 0.5 seconds (or 1000 ms) long for padding.
        silence_chunk = AudioSegment.silent(duration=padding)

        # Add the padding chunk to beginning and end of the entire chunk.
        audio_chunk = silence_chunk + chunk + silence_chunk

        # Normalize the entire chunk.
        normalized_chunk = match_target_amplitude(audio_chunk, -20.0)
        # Export the audio chunk with new bitrate.
        #print(exporting_dir + "Exporting chunk{0}.mp3.".format(i))
        normalized_chunk.export(
            ".//"+ exporting_dir + name +"{0}.wav".format(i),
            bitrate = "192k",
            format = "wav"
        )


import numpy as np
import os
import pickle
import tensorflow as tf
from tensorflow.keras import Model
from tensorflow.keras.layers import Input, Conv2D, BatchNormalization, Flatten, Dense, Reshape, Conv2DTranspose, Activation, LeakyReLU, Dropout
from tensorflow.keras.callbacks import EarlyStopping, TensorBoard
from tensorflow.keras import backend as K
from tensorflow.keras.optimizers import Adam

from model_creator.config_default import LOG_DIR

tf.compat.v1.disable_eager_execution()

class Autoencoder:
    """
    Deep Convolutionnal autoencoder with mirrored encoder and decoder components
    """

    def __init__(self, input_shape, conv_filters, conv_kernels, conv_strides, latent_space_dim, save_path = 'model'):
        """
        Initialisation of the class with the different parameters
        """

        #Shape of the input spectrogram
        self.input_shape = input_shape # [1024,256,1]
        #Shape of the differents convolutials layers
        self.conv_filters = conv_filters # [2,4,8]
        #Shape of the kernels
        self.conv_kernels = conv_kernels # [3,5,3]
        #The different strides
        self.conv_strides = conv_strides # [1,2,2]
        #The size of the latent space
        self.latent_space_dim = latent_space_dim # 256
        #Weight given to the reconstruction loss
        self.reconstruction_loss_weight = 1000000
        #The save path
        self.save_path = save_path

        #Encoder model part
        self.encoder = None
        #Decoder model part
        self.decoder = None
        #The entire model encoder + decoder
        self.model = None

        self._num_conv_layers = len(conv_filters)
        self._shape_before_bottleneck = None
        self._model_input = None

        self._build()

    def summary(self):
        '''
        Describe the network with the summary
        '''
        self.encoder.summary()
        self.decoder.summary()
        self.model.summary()

    def compile(self, learning_rate):
        """Compile the model"""
        optimizer = Adam(learning_rate  = learning_rate)

        self.model.compile(optimizer = optimizer,
                        loss = self._calculate_combined_loss, 
                        metrics = [self._calculate_reconstruction_loss, self._calculate_kl_loss])
    
    def train(self, x_train, batch_size, num_epochs, patience = 10):
        """Train the model"""
        early_stopping = EarlyStopping(monitor='loss', patience=patience, restore_best_weights=True)
        tensorboard_callback = TensorBoard(LOG_DIR, histogram_freq=1)

        history = self.model.fit(x_train,
                                x_train,
                                batch_size=batch_size,
                                epochs=num_epochs,
                                callbacks = [early_stopping, tensorboard_callback],
                                shuffle=True)
        return history
    
    def save(self, save_folder="."):
        self._create_folder_if_it_doesnt_exist(save_folder)
        self._save_parameters(save_folder)
        self._save_weights(save_folder)
    
    def load_weights(self, weights_path):
        self.model.load_weights(weights_path)

    def reconstruct(self, images):
        latent_representations = self.encoder.predict(images)
        reconstructed_images = self.decoder.predict(latent_representations)
        return reconstructed_images, latent_representations

    @classmethod
    def load(cls, save_folder="."):
        parameters_path = os.path.join(save_folder, "parameters.pkl")
        with open(parameters_path, "rb") as f:
            parameters = pickle.load(f)

        autoencoder = Autoencoder(*parameters)
        weights_path = os.path.join(save_folder, "weights.h5")
        autoencoder.load_weights(weights_path)

        return autoencoder

    def _calculate_combined_loss(self, y_target, y_predicted):
        reconstruction_loss = self._calculate_reconstruction_loss(y_target, y_predicted)
        kl_loss = self._calculate_kl_loss(y_target, y_predicted)
        combined_loss = self.reconstruction_loss_weight * reconstruction_loss + kl_loss

        return combined_loss

    def _calculate_reconstruction_loss(self, y_target, y_predicted):
        error = y_target - y_predicted
        reconstruction_loss = K.mean(K.square(error), axis = [1,2,3])
        return reconstruction_loss

    
    def _calculate_kl_loss(self, y_target, y_predicted):
        kl_loss = - 0.5 * K.sum(1 + self.log_variance - K.square(self.mu) - K.exp(self.log_variance), axis=1)
        return kl_loss

    def _create_folder_if_it_doesnt_exist(self, folder):
        if not os.path.exists(folder):
            os.mkdir(folder)

    def _save_parameters(self, save_folder):
        """
        Save the architecture of the model
        """

        parameters = [
            self.input_shape,
            self.conv_filters,
            self.conv_kernels,
            self.conv_strides,
            self.latent_space_dim
            ]

        save_path = os.path.join(save_folder, "parameters.pkl")
        with open(save_path, "wb") as f:
            pickle.dump(parameters, f)
    
    def _save_weights(self, save_folder):
        save_path = os.path.join(save_folder, "weights.h5")
        self.model.save_weights(save_path)

    def _build(self):
        self._build_encoder()
        self._build_decoder()
        self._build_autoencoder()

    def _build_autoencoder(self):
        #Input of the model
        model_input = self._model_input
        #Encoder part
        model_encoder = self.encoder(model_input)
        #Decoder part
        model_output = self.decoder(model_encoder)
        #Entire model
        self.model = Model(model_input, model_output, name="autoencoder")

    
    def _build_decoder(self):
        """
        Build the different layers for the decoder
        """

        #Decoder input
        decoder_input = self._add_decoder_input()
        #Dense layer before reshaping
        dense_layer = self._add_dense_layer(decoder_input)
        #Reshape layer to pass from 1D to 2D array
        reshape_layer = self._add_reshape_layer(dense_layer)
        #Add convolutionnal transpose blocks
        conv_transpose_layer = self._add_conv_transpose_layers(reshape_layer)
        #Add decoder output
        decoder_output = self._add_decoder_output(conv_transpose_layer)
        #Model final composition
        self.decoder = Model(decoder_input, decoder_output, name = 'decoder')
    

    def _add_decoder_input(self):
        return Input(shape=(self.latent_space_dim,), dtype=tf.float32, name="decoder_input")

    def _add_dense_layer(self, decoder_input):
        """
        Te different blocs to decompress the botleneck
        """

        num_neurons = np.prod(self._shape_before_bottleneck) # [1, 2, 4] -> 8
        dense_layer = Dense(self.latent_space_dim * 2, name="decoder_dense_1", activation= 'relu')(decoder_input)
        dense_layer = Dense(self.latent_space_dim * 4, name="decoder_dense_2", activation= 'relu')(dense_layer)
        dense_layer = Dense(num_neurons, name="decoder_dense_3", activation = 'relu')(dense_layer)
        return dense_layer
    
    def _add_reshape_layer(self, dense_layer):
        return Reshape(self._shape_before_bottleneck)(dense_layer)

    def _add_conv_transpose_layers(self, x):
        """Add conv transpose blocks."""
        for layer_index in reversed(range(1, self._num_conv_layers)):
            x = self._add_conv_transpose_layer(layer_index, x)
        return x

    
    def _add_conv_transpose_layer(self, layer_index, x):
        layer_num = self._num_conv_layers - layer_index
        conv_transpose_layer = Conv2DTranspose(
            filters = self.conv_filters[layer_index], # [24, 24, 1]
            kernel_size = self.conv_kernels[layer_index],
            strides = self.conv_strides[layer_index],
            padding = "same",
            name=f"decoder_conv_transpose_layer_{layer_num}"
        )
        x = conv_transpose_layer(x)
        x = LeakyReLU(name=f"decoder_leaky_relu_{layer_num}")(x)
        x = BatchNormalization(name =f"decoder_bn_{layer_num}")(x)
        return x

    def _add_decoder_output(self, x):
        conv_transpose_layer = Conv2DTranspose(
            filters = 1, # [24, 24, 1]
            kernel_size = self.conv_kernels[0],
            strides = self.conv_strides[0],
            padding = "same",
            name=f"decoder_conv_transpose_layer_output{self._num_conv_layers}"
        )

        x = conv_transpose_layer(x)
        output_layer = Activation("sigmoid", name="sigmoid_layer")(x)
        return output_layer

    def _build_encoder(self):
        """
        Build the different layers of the encoder
        """

        #Input
        encoder_input = self._add_encoder_input()
        #Convolution layer
        conv_layer = self._add_conv_layers(encoder_input)
        #Bottleneck
        bottleneck = self._add_bottleneck(conv_layer)
        self._model_input = encoder_input
        self.encoder = Model(encoder_input, bottleneck, name = "encoder")

    def _add_encoder_input(self):
        return Input(shape=self.input_shape, name = "encoder_input")
    
    def _add_conv_layers(self, encoder_input):
        x = encoder_input
        for layer_index in range(self._num_conv_layers):
            x = self._add_conv_layer(layer_index, x)
        return x
    
    def _add_conv_layer(self, layer_index, x):
        """
        Define the convolutionnal block
        """
        layer_number = layer_index + 1
        conv_layer = Conv2D(
            filters = self.conv_filters[layer_index],
            kernel_size = self.conv_kernels[layer_index],
            strides = self.conv_strides[layer_index],
            padding = "same",
            name = f"encoder_conv_layer_{layer_index + 1}"
        )

        x = conv_layer(x)
        x = LeakyReLU(name=f"encoder_leakyrelu_{layer_number}")(x)
        x = BatchNormalization(name=f"encoder_bn_{layer_number}")(x)
        return x
    
    def _add_bottleneck(self, x):
        """Flatten data and add bottleneck with Gaussian sampling (Dense layer)."""
        self._shape_before_bottleneck = K.int_shape(x)[1:] # [7, 7, 32]
        x = Flatten()(x)
        x = Dense(self.latent_space_dim * 4, name = "first_reduction", activation="relu")(x)
        x = Dropout(0.3)(x)
        x = Dense(self.latent_space_dim * 2, name = "second_reduction", activation="relu")(x)
        x = Dropout(0.3)(x)
        self.mu = Dense(self.latent_space_dim, name="mu")(x)
        self.log_variance = Dense(self.latent_space_dim, name="log_variance")(x)

        def sample_point_from_normal_distribution(args):
            mu, log_variance = args
            epsilon = K.random_normal(shape=K.shape(self.mu), mean=0., stddev=1.)
            sampled_point = mu + K.exp(log_variance / 2) * epsilon

            return sampled_point

        x = tf.keras.layers.Lambda(sample_point_from_normal_distribution, name="encoder_output")([self.mu, self.log_variance])
        return x


"""
1- load a file
2- pad the signal (if necessary)
3- extracting log spectrogram from signal
4- normalise spectrogram
5- save the normalised spectrogram
PreprocessingPipeline
"""
import os
import pickle
import librosa
import numpy as np
import zipfile

class Data_Recup:
    """
    Get the data of birds from kaggle 
    """

    def __init__(self, file_key_path : str):
        self.file_key_path = file_key_path

    def get_songs(self) -> None:
        """
        Download the birds songs
        """

        os.system('mkdir ~/.kaggle')
        os.system('cp ' + self.file_key_path + ' ~/.kaggle/')
        os.system('chmod 600 ~/.kaggle/kaggle.json')
        os.system('kaggle datasets download -d monogenea/birdsongs-from-europe')
        
        with zipfile.ZipFile('birdsongs-from-europe.zip', 'r') as zip_ref:
            zip_ref.extractall(".")

class Loader:
    """Loader is responsible for loading an audio file."""

    def __init__(self, sample_rate:int, duration:int, mono):
        self.sample_rate = sample_rate
        self.duration = duration
        self.mono = mono

    def load(self, file_path):
        signal = librosa.load(file_path,
                              sr=self.sample_rate,
                              duration=self.duration,
                              mono=self.mono)[0]
        return signal


class Padder:
    """Padder is responsible to apply padding to an array."""

    def __init__(self, mode="constant"):
        self.mode = mode

    def left_pad(self, array, num_missing_items):
        padded_array = np.pad(array,
                              (num_missing_items, 0),
                              mode=self.mode)
        return padded_array

    def right_pad(self, array, num_missing_items):
        padded_array = np.pad(array,
                              (0, num_missing_items),
                              mode=self.mode)
        return padded_array


class LogSpectrogramExtractor:
    """LogSpectrogramExtractor extracts log spectrograms (in dB) from a
    time-series signal.
    """

    def __init__(self, frame_size:int, hop_length:int):
        self.frame_size = frame_size
        self.hop_length = hop_length

    def extract(self, signal):
        stft = librosa.stft(signal,
                            n_fft=self.frame_size,
                            hop_length=self.hop_length)[:-1]
        spectrogram = np.abs(stft)
        log_spectrogram = librosa.amplitude_to_db(spectrogram)
        return log_spectrogram


class MinMaxNormaliser:
    """MinMaxNormaliser applies min max normalisation to an array."""

    def __init__(self, min_val, max_val):
        self.min = min_val
        self.max = max_val

    def normalise(self, array):
        norm_array = (array - array.min()) / (array.max() - array.min())
        norm_array = norm_array * (self.max - self.min) + self.min
        return norm_array

    def denormalise(self, norm_array, original_min, original_max):
        array = (norm_array - self.min) / (self.max - self.min)
        array = array * (original_max - original_min) + original_min
        return array


class Saver:
    """saver is responsible to save features, and the min max values."""

    def __init__(self, feature_save_dir, min_max_values_save_dir):
        self.feature_save_dir = feature_save_dir
        self.min_max_values_save_dir = min_max_values_save_dir

    def save_feature(self, feature, file_path):
        save_path = self._generate_save_path(file_path)
        np.save(save_path, feature)
        return save_path

    def save_min_max_values(self, min_max_values):
        save_path = os.path.join(self.min_max_values_save_dir,
                                 "min_max_values.pkl")
        self._save(min_max_values, save_path)

    @staticmethod
    def _save(data, save_path):
        with open(save_path, "wb") as f:
            pickle.dump(data, f)

    def _generate_save_path(self, file_path):
        file_name = os.path.split(file_path)[1]
        save_path = os.path.join(self.feature_save_dir, file_name + ".npy")
        return save_path


class PreprocessingPipeline:
    """PreprocessingPipeline processes audio files in a directory, applying
    the following steps to each file:
        1- load a file
        2- pad the signal (if necessary)
        3- extracting log spectrogram from signal
        4- normalise spectrogram
        5- save the normalised spectrogram
    Storing the min max values for all the log spectrograms.
    """

    def __init__(self):
        self.padder = None
        self.extractor = None
        self.normaliser = None
        self.saver = None
        self.min_max_values = {}
        self._loader = None
        self._num_expected_samples = None

    @property
    def loader(self):
        return self._loader

    @loader.setter
    def loader(self, loader):
        self._loader = loader
        self._num_expected_samples = int(loader.sample_rate * loader.duration)

    def process(self, audio_files_dir):
        for root, _, files in os.walk(audio_files_dir):
            for file in files:
                file_path = os.path.join(root, file)
                self._process_file(file_path)
                print(f"Processed file {file_path}")
        self.saver.save_min_max_values(self.min_max_values)

    def process_applicatif(self, audio_files_dir):
        spectrograms = []
        for root, _, files in os.walk(audio_files_dir):
            for file in files:
                file_path = os.path.join(root, file)
                spectrograms.append(self._process_file(file_path, write_file=False))
                print(f"Processed file {file_path}")
        return np.array(spectrograms)

    def _process_file(self, file_path, write_file = True):
        signal = self.loader.load(file_path)
        if self._is_padding_necessary(signal):
            signal = self._apply_padding(signal)
        feature = self.extractor.extract(signal)
        norm_feature = self.normaliser.normalise(feature)
        if write_file:
            save_path = self.saver.save_feature(norm_feature, file_path)
            self._store_min_max_value(save_path, feature.min(), feature.max())
        else:
            return norm_feature

    def _is_padding_necessary(self, signal):
        if len(signal) < self._num_expected_samples:
            return True
        return False

    def _apply_padding(self, signal):
        num_missing_samples = self._num_expected_samples - len(signal)
        padded_signal = self.padder.right_pad(signal, num_missing_samples)
        return padded_signal

    def _store_min_max_value(self, save_path, min_val, max_val):
        self.min_max_values[save_path] = {
            "min": min_val,
            "max": max_val
        }


from unittest.mock import DEFAULT


import datetime

FRAME_SIZE = 2048
HOP_LENGTH = 512
DURATION = 2.6
SAMPLE_RATE = 22050
MONO = True

#Default compile model
DEFAULT_LEARNING_RATE = 0.00015
DEFAULT_BATCH_SIZE = 8
DEFAULT_EPOCHS = 150
DEFAULT_PATIENCE = 10

MODEL_PATH = "./model/"
MODEL_PATH_CLASSIQUE = "./model/classique_model.h5"
DEFAULT_MODEL_TUNING_TRIAL = 50

LOG_DIR = "logs/" + datetime.datetime.now().strftime("%Y%m%d-%H%M%S")


from setuptools import find_packages, setup

setup(
    name='model_creator_bird_sing_v2',
    packages=['model_creator/'],
    version='0.2.0',
    description='Autoencoder singing',
    author='VIRGAUX Pierre',
    license='MIT',
    install_requires=['tensorflow==2.3.1','mutagen','pydub','ray', 'librosa', 'keras-tuner', 'wandb'],
)